"""

"""
# Partial imports
from abc import ABC

# Imports from internal modules
from .instance_solution import InstanceSolutionCore


class SolutionCore(InstanceSolutionCore, ABC):
    """
    The solution template.
    """

    pass
